/**
 * JMSwingTextPane, a Swing TextPane that has been extended to
 * allow programmers to turn antialiasing on and off easily.
 * $Id: JMSwingTextPane.java,v 1.1 2004/07/28 02:43:42 jeffnik Exp $
*/

/* JMSwingTextPane, a antialiasing TextPane
 * Copyright (C) 2004 Jeff Robinson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2, as published by the Free Software Foundation; 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package anecho.gui;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;

import javax.swing.JTextPane;
import javax.swing.text.StyledDocument;

public class JMSwingTextPane extends JTextPane {
    // java.util.Map hints = new java.util.HashMap();
    // {
    // hints.put( RenderingHints.KEY_ANTIALIASING,
    // RenderingHints.VALUE_ANTIALIAS_OFF);
    // }

    boolean antiAlias = false;
    private static final boolean DEBUG = false;

    public JMSwingTextPane() {
        super();
    }

    public JMSwingTextPane(StyledDocument doc) {
        super(doc);
    }

    /** Set the status of our AntiAliasing to either be
     * @param true - antialiased (slower but pretty)
     * @param false - not anti-alised (faster but not as pretty)
     */
    public synchronized void setAntiAliasing(boolean status) {
        antiAlias = status;

        if (DEBUG) {
            System.err.println("JMSwingTextPane.setAntiAliasing() Setting antiAliasing to: " + status);
        }

        // Try to ensure that our display gets updated to show the new look
        //repaint();
        // paint();
        invalidate();
        validate();
        repaint();

        /*
        if (hints.containsKey(RenderingHints)) {
            hints.remove( RenderingHints.KEY_ANTIALIASING);
        }

        if (status) {
            hints.put( RenderingHints.KEY_ANTIALIASING,
                       RenderingHints.VALUE_ANTIALIAS_ON);
        } else {
            hints.put( RenderingHints.KEY_ANTIALIASING,
                       RenderingHints.VALUE_ANTIALIAS_OFF);
                       }
                       */
    }

    /** Over-ridden paintComponent proceedure to set the Rendering Hints
     */
    public void paintComponent( Graphics g ) {

        Graphics2D g2 = (Graphics2D) g;

        if (antiAlias) {
            if (DEBUG) {
                System.err.print("+");
            }
            g2.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON );
        } else {
            if (DEBUG) {
                System.err.print("-");
            }
            g2.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_OFF );
            g2.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF );
        }


        // g2.addRenderingHints( hints );
        super.paintComponent( g );
    }

}